/*
 * @(#) Prolog.cs
 * Prolog engine & runtime environment
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

// import java.lang.*;
// import java.util.*;
// import java.io.*;

namespace JJC.Psharp.Lang {

	using JJC.Psharp.Lang.Resource;

	using System;
	using System.Collections;
	using System.IO;
	using System.Text;

	public sealed class Prolog {
    
		public readonly long runtime0 = System.DateTime.Now.Ticks / 10000; //System.currentTimeMillis();
		public long old_runtime = 0;

    
		public static readonly SymbolTerm Nil = SymbolTerm.makeSymbol("[]");

    
		public readonly PushbackReader user_input  
			= new PushbackReader( System.Console.In );
		public readonly System.IO.TextWriter user_output 
			= System.Console.Out;
		public readonly System.IO.TextWriter user_error 
			= System.Console.Error;
    
		public PushbackReader current_input; //= user_input;
		public TextWriter current_output; //= user_output;

    
		public readonly Term[] aregs = new Term[255];    
		public Predicate cont = null;                 
		public readonly CpfStack stack = new CpfStack(); // Choice point frame
		public readonly Trail trail = new Trail();       // Trail
		public int B0; // = stack.top;                  // Cut

    
		public readonly Hashtable user_table = new Hashtable();
		public Hashtable current_table = new Hashtable();

		public long timeStamp = System.Int64.MinValue;

		public readonly ResourceTable res = new ResourceTable();
		public int L =  1;         // $B8=:_>CHq2DG=$J%j%=!<%9$N%l%Y%k(B
		public int U = -1;         // $B>CHq8e$K%;%C%H$5$l$kCM(B
		public int R0; // = res.top; // $B4XO"%j%9%H:n@.$K;HMQ(B(begin_imp,more_imp)
		public Term R1;            // $B>CHq2DG=$J(Bindex$B$NA4%j%9%H(B(by hash)
		public Term R2;            // $B>CHq2DG=$J(Bindex$B$N%j%9%H(B($B%j%=!<%9DI2C;~$KBh(B1$B0z?t$,JQ?t(B)

		public Prolog() { 
			R0 = res.top;
			B0 = stack.top;
			current_input = user_input;
			current_output = user_output;
			res.setEngine(this);
			trail.setEngine(this);
			init();
		}

		public void init() { 
			Term[] noarg = {};
			stack.create(noarg, null);
			stack.setTR(trail.top);
			stack.setTimeStamp(++timeStamp);
			stack.setBP(null);    
			stack.setB0(B0);
			stack.setR(res.top);
			stack.setL(L);
			stack.setU(U);
		}

		public VariableTerm makeVariable(){
			return new VariableTerm(timeStamp);
		}

		// $B3F<o%G!<%?%(%j%"$N>pJs<hF@(B
		public int stack_top(){ return stack.top; }
		public int stack_max(){ return stack.max(); }
		public int trail_top(){ return trail.top; }
		public int trail_max(){ return trail.max(); }
		public int resTop()   { return res.top; } 

		// $B9`$N%3%T!<(B(copy_term/2$B$G;HMQ(B)
		public Hashtable varTable = new Hashtable();
		public Term copyTerm(Term t){
			t = t.dereference();
			varTable.Clear();
			return t.copyTerm(this);
		}

		// $B%F!<%V%k!"F~=PNO$N(Bcurrent$BCM$N@_Dj(B
		public void setTable(Hashtable table){
			current_table = table;
		}
		public void setInput(PushbackReader cin){
			current_input = cin;
		}
		public void setOutput(TextWriter cout){
			current_output = cout;
		}

		public Predicate abort(string s){ throw new AbortException(s); }

		public Predicate fail() {
			if ( stack.empty() ) 
				return null;
			B0 = stack.getB0();     // B0$B$r85$KLa$9(B
			return stack.getBP();   // next clause$B$r<B9T(B
		}

		public void setB0()    { B0 = stack.top; }  // cut point$B$N%;%C%H(B
		public void neckCut()  { stack.cut(B0); }     // neck cut$B$N<B9T(B
		public void cut(int i) { stack.cut(i); }      // $BIaDL$N(Bcut$B$N<B9T(B

		public void restore(){
			Term[] args = stack.getArgs();
			int i = args.Length;
			System.Array.Copy(args, 0, aregs, 1, i);
			cont = stack.getCont();
		}

		// switch_on_term
		public Predicate switch_on_term(Predicate var, 
			Predicate num, 
			Predicate con, 
			Predicate str, 
			Predicate list) {
			Term arg1 = aregs[1].dereference();
			if ( arg1.isVariable() )  return var;
			if ( arg1.isInteger() )   return num;
			if ( arg1.isSymbol() )    return con;
			if ( arg1.isStructure() ) return str;
			if ( arg1.isList() )      return list;
			return fail();
		}

		// try
		public Predicate jtry(Predicate p, Predicate next){
			int i = p.arity();
			Term[] args = new Term[i];
			System.Array.Copy(aregs, 1, args, 0, i);
			stack.create(args, cont);
			stack.setTR(trail.top);
			stack.setTimeStamp(++timeStamp);
			stack.setBP(next);
			stack.setB0(B0);
			// for LLP
			stack.setR(res.top);
			stack.setL(L);
			stack.setU(U);
			return p;
		}

		// retry
		public Predicate retry(Predicate p, Predicate next){
			restore();
			trail.unwind(stack.getTR());
			stack.setBP(next);
			// for LLP
			res.unwind(stack.getR());
			L = stack.getL();
			U = stack.getU();
			return p;
		}

		// trust
		public Predicate trust(Predicate p){
			restore();
			trail.unwind(stack.getTR());
			// for LLP
			res.unwind(stack.getR());
			L = stack.getL();
			U = stack.getU();
			stack.delete();
			return p;
		}


		//----------------------------------------------------------------- 
		//                        $B%j%=!<%9=hM}4XO"(B 
		//----------------------------------------------------------------- 

		// $B4XO"%j%=!<%9$NHO0O$N@8@.$K;HMQ(B
		public void setR0(){ R0 = res.top; }

		// $B3F<o%j%=!<%9%l%Y%k$X$NA`:n(B
		public void decrementU(){ U--; }
		public void incrementU(){ U++; }
		public void decrementL(){ L--; }
		public void incrementL(){ L++; }


		public void lookUpHash(SymbolTerm functor){
			res.setR1andR2(functor);
		}

		public bool pickupResource(SymbolTerm sym, int N){
			Term t;
			for (; R1.isList(); R1 = ((ListTerm)R1).cdr){
				t = ((ListTerm)R1).car;
				if ( !t.isResIndex() ){
					System.Console.WriteLine("error in pickupResource\n");
					return false;
				}
				if ( res.consumable(sym, ((ResIndexTerm)t).getIndex()) ){
					aregs[N] = t;
					R1 = ((ListTerm)R1).cdr;
					return true;
				}
			}
			for (; R2.isList(); R2 = ((ListTerm)R2).cdr){
				t = ((ListTerm)R2).car;
				if ( !t.isResIndex() ){
					System.Console.WriteLine("error in pickupResource\n");
					return false;
				}
				if ( res.consumable(sym, ((ResIndexTerm)t).getIndex()) ){
					aregs[N] = t;
					R2 = ((ListTerm)R2).cdr;
					return true;
				}
			}
			return false;
		}

		// Linear$B%j%=!<%9$NDI2C(B
		public void addRes(Term head, ClosureTerm closure){
			res.addRes(head, closure);
		}
		// Exponential$B%j%=!<%9$NDI2C(B
		public void addExpRes(Term head, ClosureTerm closure){
			res.addExpRes(head, closure);
		}
		// R1, R2$B$N%;%C%H(B
		public void setRelatedResField(){
			res.setRelatedResField();
		}
		public void midWith(){ 
			res.midWith(); 
		}
		public bool endWith(){ 
			return res.endWith(); 
		}
		public bool endImp(ResIndexTerm i, ResIndexTerm j){
			return res.endImp(i,j);
		}
		public void endExpImp(ResIndexTerm i, ResIndexTerm j){
			res.endExpImp(i, j);
		}

		// try resource
		public Predicate tryResource(Predicate p, Predicate next){
			int i;  /*  num_of_args */
			Term[] args;

			i = p.arity();
			args = new Term[i];
			System.Array.Copy(aregs, 1, args, 0, i);
			stack.create(args, cont);
			stack.setTR(trail.top);
			stack.setTimeStamp(++timeStamp); //added
			stack.setBP(next);
			stack.setB0(B0);
			stack.setR1(R1);
			stack.setR2(R2);
			stack.setR(res.top);
			stack.setL(L);
			stack.setU(U);
			return p;
		}

		// choice point frame$B$+$i%l%8%9%?$NCM$rLa$9(B
		public void restoreResource(){
			restore();
			trail.unwind(stack.getTR());
			R1 = stack.getR1();
			R2 = stack.getR2();
			res.unwind(stack.getR());
			L = stack.getL();
			U = stack.getU();
		}

		// R1,R2$B$r8=:_$NCM$K99?7$7!"(BBP$B$b%;%C%H$7D>$9(B
		public Predicate retryResource(Predicate p, Predicate next){
			stack.setR1(R1);
			stack.setR2(R2);
			stack.setBP(next);
			return p;
		}

		// choice point frame$B$r:o=|(B
		public Predicate trustResource(Predicate p){
			stack.delete();
			return p;
		}

		// $B%j%=!<%9$K>CHq%^!<%/$rIU$1$k(B
		public ClosureTerm consume(int N){
			if ( ! aregs[N].isResIndex() )
				System.Console.WriteLine("error in consume/1 in Prolog.cs");
			int index = ((ResIndexTerm)(aregs[N])).getIndex();
			res.consume(index);
			return res.getClosureAt(index);
		}

		// closure$B$N<B9T(B
		public Predicate executeClosure(ClosureTerm closure){
			Predicate code = closure.code;
			Term[] vars = closure.freeVars;
			int len = closure.length;
			if ( len == 0 ) 
				return code;
			System.Array.Copy(vars, 0, aregs, code.arity()+1, len);
			return code;
		}

		// $B%*%W%F%#%^%$%:MQ(B
		public bool hasMoreResource(){
			Term t;
			for (; R1.isList(); R1 = ((ListTerm)R1).cdr){
				t = ((ListTerm)R1).car;
				if ( !t.isResIndex() ){
					System.Console.WriteLine("error in noResource\n");
					return true;
				}
				if ( res.consumable(((ResIndexTerm)t).getIndex()) ){
					break;
				}
			}
			for (; R2.isList(); R2 = ((ListTerm)R2).cdr){
				t = ((ListTerm)R2).car;
				if ( !t.isResIndex() ){
					System.Console.WriteLine("error in noResource\n");
					return true;
				}
				if ( res.consumable(((ResIndexTerm)t).getIndex()) ) {
					break;
				}
			}
			if ( R1.isNil() ){
				R1 = R2;
				R2 = Nil;
			}
			return R1.isList();
		}

		// top$B%4!<%k$N<B9T(B
		public void executeTop(){ 
			res.executeTop(); 
		}

		//---------------------------------------------------------------------
		//                            $B;;=Q1i;;(B
		//---------------------------------------------------------------------
		public SymbolTerm SYM_plus_1    = SymbolTerm.makeSymbol("+", 1);
		public SymbolTerm SYM_plus_2    = SymbolTerm.makeSymbol("+", 2);
		public SymbolTerm SYM_minus_1   = SymbolTerm.makeSymbol("-", 1);
		public SymbolTerm SYM_minus_2   = SymbolTerm.makeSymbol("-", 2);
		public SymbolTerm SYM_times_2   = SymbolTerm.makeSymbol("*", 2);
		public SymbolTerm SYM_floatquot_2 = SymbolTerm.makeSymbol("/", 2);
		public SymbolTerm SYM_intquot_2 = SymbolTerm.makeSymbol("//", 2);
		public SymbolTerm SYM_mod_2     = SymbolTerm.makeSymbol("mod", 2);
		public SymbolTerm SYM_max_2     = SymbolTerm.makeSymbol("max", 2);
		public SymbolTerm SYM_min_2     = SymbolTerm.makeSymbol("min", 2);
		public SymbolTerm SYM_rshift_2  = SymbolTerm.makeSymbol(">>", 2);
		public SymbolTerm SYM_lshift_2  = SymbolTerm.makeSymbol("<<", 2);
		public SymbolTerm SYM_bitconj_2 = SymbolTerm.makeSymbol("/\\", 2);
		public SymbolTerm SYM_bitdisj_2 = SymbolTerm.makeSymbol("\\/", 2);
		public SymbolTerm SYM_bitneg_1  = SymbolTerm.makeSymbol("\\", 1);
		public SymbolTerm SYM_bitxor_2  = SymbolTerm.makeSymbol("#", 2);
		public SymbolTerm SYM_abs_1     = SymbolTerm.makeSymbol("abs", 1);
		public SymbolTerm SYM_int_1     = SymbolTerm.makeSymbol("integer", 1);

		//    public int evalTerm(Term t) {
		public int evalTerm(Term t) {
			t = t.dereference();
			if ( t.isInteger() ) 
				return ((IntegerTerm) t).value();
			if ( t.isList() )
				return evalTerm(((ListTerm)t).car);
			if ( t.isStructure() ){
				StructureTerm t1 = (StructureTerm) t; 
				SymbolTerm func = t1.functor;
				Term[] args = t1.args;
				int arity = t1.arity;

				if( func == SYM_plus_1 )
					return evalTerm(args[0]);
				else if( func == SYM_plus_2 )
					return evalTerm(args[0]) + evalTerm(args[1]);
				else if ( func == SYM_minus_1 )
					return - evalTerm(args[0]);
				else if ( func == SYM_minus_2 )
					return evalTerm(args[0]) - evalTerm(args[1]);
				else if( func == SYM_times_2 )   
					return evalTerm(args[0]) * evalTerm(args[1]);
				else if( func == SYM_intquot_2 )
					return evalTerm(args[0]) / evalTerm(args[1]);
				else if( func == SYM_floatquot_2 )
					return evalTerm(args[0]) / evalTerm(args[1]);
				else if( func == SYM_mod_2 )   
					return evalTerm(args[0]) % evalTerm(args[1]);
				else if ( func == SYM_abs_1 )
					return Math.Abs(evalTerm(args[0]));
				else if( func == SYM_max_2 )   
					return Math.Max(evalTerm(args[0]), evalTerm(args[1]));
				else if( func == SYM_min_2 )   
					return Math.Min(evalTerm(args[0]), evalTerm(args[1]));
				else if( func == SYM_rshift_2 )   
					return evalTerm(args[0]) >> evalTerm(args[1]);
				else if( func == SYM_lshift_2 )   
					return evalTerm(args[0]) << evalTerm(args[1]);
				else if( func == SYM_bitneg_1 )
					return ~ evalTerm(args[0]);
				else if( func == SYM_bitconj_2 )   
					return evalTerm(args[0]) & evalTerm(args[1]);
				else if( func == SYM_bitdisj_2 )   
					return evalTerm(args[0]) | evalTerm(args[1]);
				else if( func == SYM_bitxor_2 )   
					return evalTerm(args[0]) ^ evalTerm(args[1]);
				else if ( func == SYM_int_1 )
					return (int)evalTerm(args[0]);
			}
			//	System.Console.WriteLine("The term " + t + " is not arithmetic expression.");
			throw (new ArithException());
		}

		//---------------------------------------------------------------------
		//                           $B9`$NHf3S(B
		//---------------------------------------------------------------------
		public static int EQUAL  = 0;
		public static int BEFORE = -1;
		public static int AFTER  = 1;

		// compare function
		// Variables < Integers < Atoms < Compound terms(containing lists)
		public static int compareTerm(Term t1, Term t2) {
			Term a1 = t1.dereference();
			Term a2 = t2.dereference();

			if ( a1 == a2 ) return EQUAL;
			if ( a1.isVariable() ){
				if ( !a2.isVariable() ) return BEFORE;
				//	    long i = ((VariableTerm)a1).timeStamp();
				//	    long j = ((VariableTerm)a2).timeStamp();
				long i = ((VariableTerm)a1).getID();
				long j = ((VariableTerm)a2).getID();
				if ( i < j ) return BEFORE;
				return AFTER;
			}
			if ( a1.isInteger() ){
				if ( a2.isVariable() ) return AFTER;
				if ( !a2.isInteger() ) return BEFORE;
				int i = ((IntegerTerm)a1).value();
				int j = ((IntegerTerm)a2).value();
				if ( i < j )  return BEFORE;
				if ( i == j ) return EQUAL;
				return AFTER;
			}
			if ( a1.isSymbol() ){
				if (a2.isVariable() || a2.isInteger())
					return AFTER;
				if ( !a2.isSymbol() ) return BEFORE;
				string i = ((SymbolTerm)a1).name;
				string j = ((SymbolTerm)a2).name;
				return i.CompareTo(j) ;
			}
			if ( a1.isList() ){
				if (a2.isVariable() || a2.isInteger() || a2.isSymbol())
					return AFTER;
				if ( a2.isStructure() ){
					if (  ((StructureTerm)a2).arity >= 2 ) 
						return BEFORE;
					return AFTER;
				}
				Term car1 = ((ListTerm)a1).car;
				Term car2 = ((ListTerm)a2).car;
				int result = compareTerm(car1, car2);
				if ( result != 0 ) return result;
				Term cdr1 = ((ListTerm)a1).cdr;
				Term cdr2 = ((ListTerm)a2).cdr;
				return compareTerm(cdr1, cdr2);
			}
			if ( a1.isStructure() ){
				if ( a2.isVariable() || a2.isInteger() || a2.isSymbol() )
					return AFTER;

				SymbolTerm f1 = ((StructureTerm)a1).functor;
				int arity1 = ((StructureTerm)a1).arity;
				Term[] ag1 = ((StructureTerm)a1).args;
				if ( a2.isList() ){
					if(arity1 < 2) return BEFORE;
					return AFTER;
				}
				SymbolTerm f2 = ((StructureTerm)a2).functor;
				int arity2 = ((StructureTerm)a2).arity;
				Term[] ag2 = ((StructureTerm)a2).args;

				if ( arity1 < arity2 ) return BEFORE ;
				if ( arity1 > arity2 ) return AFTER ;		
				int ct = f1.ToString().CompareTo(f2.ToString());
				if ( ct != 0 ) return ct;
				for (int i=0 ; i<arity1 ; i++) {
					Term agi = ag1[i].dereference();
					Term agj = ag2[i].dereference();
					ct = compareTerm(agi, agj);
					if ( ct != 0 ) return ct;
				}
				return EQUAL;
			}
			return AFTER;
		}
	}

}